#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <stralloc.h>
#include <time.h>
#include <unistd.h>
#include <afblib/mpx_session.h>
#include "mastermind-service.h"

typedef struct mastermind_session {
   stralloc name;
   stralloc param; /* input buffer for optional parameter, temporarily used */
   session* s;
   struct mastermind_session* prev;
   struct mastermind_session* next;
} mastermind_session;

#define MAX_COLOURS 9 /* just digits, please */
#define MAX_LENGTH 20

typedef struct {
   unsigned int len; /* length of the code */
   unsigned int colours; /* number of different colours
      (or digits) used in the code */
   bool unique; /* are the colours (or digits) unique in the code? */
   unsigned int guesses; /* number of guesses so far */
   bool running; /* has the game been initialized and not yet guessed? */
   char code[MAX_LENGTH + 1]; /* actual code */
   unsigned int colourset; /* bitset of all colours used in the code */
   unsigned int count[MAX_COLOURS]; /* number of occurrences of each colour */
} Game;

typedef struct {
   Game game;
   stralloc status; /* textual representation of the actual status */
   stralloc line; /* temporarily used for the next status line */
   stralloc guess; /* input buffer, temporarily used */
   mastermind_session* ring; /* ring of currently running sessions */
} mastermind;

/* create a new code for a new game using the
   parameters len, colours, and unique */
void init_game(Game* game) {
   for (unsigned int i = 0; i < MAX_COLOURS; ++i) {
      game->count[i] = 0;
   }
   unsigned int set = 0; /* bitset of colours */
   for (unsigned int i = 0; i < game->len; ++i) {
      unsigned int colour; unsigned int mask;
      do {
	 colour = rand() % game->colours;
	 mask = (1 << colour);
      } while (game->unique && (set & mask));
      set |= mask;
      ++game->count[colour];
      game->code[i] = colour + '0' + 1;
   }
   game->code[game->len] = 0;
   game->guesses = 0;
   game->colourset = set;
   game->running = true;
}

void start_game(mastermind* mm) {
   init_game(&mm->game);
   mm->status.len = 0;
   stralloc_copys(&mm->status, "Parameters: len = ");
   stralloc_catint(&mm->status, mm->game.len);
   stralloc_cats(&mm->status, ", colours = ");
   stralloc_catint(&mm->status, mm->game.colours);
   if (mm->game.unique) {
      stralloc_cats(&mm->status, ", unique = true");
   } else {
      stralloc_cats(&mm->status, ", unique = false");
   }
}

/* macros used for bitset operations for colours */
#define COLOUR(colour) \
   ((colour)-'0'-1)
#define IN_SET(colour, set) \
   ((1<<COLOUR(colour)) & set)

/* evaluate a guess by setting *black_counter and *white_counter,
   if the guess appears irregular, false is returned */
bool evaluate_guess(Game* game, stralloc* guess,
      unsigned int* black_counter, unsigned int* white_counter) {
   if (game->len != guess->len) return false;
   unsigned black = 0; unsigned int white = 0;
   if (game->unique) {
      for (unsigned int i = 0; i < game->len; ++i) {
	 if (guess->s[i] == game->code[i]) {
	    ++black;
	 } else if (guess->s[i] == '0' ||
	       COLOUR(guess->s[i]) >= game->colours) {
	    /* colour used which cannot appear in the code */
	    return false;
	 } else if (IN_SET(guess->s[i], game->colourset)) {
	    ++white;
	 }
      }
   } else {
      unsigned int count[MAX_COLOURS] = {0};
      for (unsigned int i = 0; i < game->len; ++i) {
	 if (guess->s[i] == '0' || COLOUR(guess->s[i]) >= game->colours) {
	    /* colour used which cannot appear in the code */
	    return false;
	 }
	 count[COLOUR(guess->s[i])]++;
	 if (guess->s[i] == game->code[i]) {
	    ++black;
	 }
      }
      for (unsigned int i = 0; i < game->colours; ++i) {
	 if (count[i] <= game->count[i]) {
	    white += count[i];
	 } else {
	    white += game->count[i];
	 }
      }
      white -= black;
   }
   ++game->guesses;
   *black_counter = black;
   *white_counter = white;
   return true;
}

/* convert decimal sequence of digits to an integer value */
unsigned int get_value(stralloc* param) {
   unsigned int val = 0;
   for (unsigned int i = 0; i < param->len; ++i) {
      val = val * 10 + param->s[i] - '0';
   }
   return val;
}

static void broadcast(mastermind* mm, mastermind_session* sender,
      stralloc* message) {
   mastermind_session* ms = mm->ring;
   if (ms) {
      do {
	 if (ms->name.len > 0) { /* registered users only */
	    if (sender) {
	       mpx_session_printf(ms->s, "%.*s: %.*s\r\n",
		  sender->name.len, sender->name.s,
		  message->len, message->s);
	    } else {
	       mpx_session_printf(ms->s, "%.*s\r\n",
		  message->len, message->s);
	    }
	 }
	 ms = ms->next;
      } while (ms != mm->ring);
   }
}

static void broadcasts(mastermind* mm, mastermind_session* sender,
      char* message) {
   mastermind_session* ms = mm->ring;
   if (ms) {
      do {
	 if (ms->name.len > 0) { /* registered users only */
	    if (sender) {
	       mpx_session_printf(ms->s, "%.*s: %s\r\n",
		  sender->name.len, sender->name.s, message);
	    } else {
	       mpx_session_printf(ms->s, "%s\r\n",
		  message);
	    }
	 }
	 ms = ms->next;
      } while (ms != mm->ring);
   }
}

/* reveal the code if the game was still running
   and finish the game */
void finish_game(mastermind* mm, mastermind_session* ms) {
   if (mm->game.running && mm->game.guesses > 0) {
      stralloc message = {0};
      stralloc_copys(&message, "gives up after ");
      stralloc_catint(&message, mm->game.guesses);
      stralloc_cats(&message, " fruitless attempts to guess ");
      stralloc_cats(&message, mm->game.code);
      broadcast(mm, ms, &message);
      stralloc_free(&message);
      mm->game.guesses = 0;
   }
   mm->game.running = false;
}

static void open_handler(session* s) {
   mastermind_session* ms = calloc(1, sizeof(mastermind_session));
   if (!ms) {
      close_session(s); return;
   }
   s->handle = ms;
   mastermind* mm = (mastermind*) s->global_handle;
   ms->s = s;
   if (mm->ring) {
      ms->next = mm->ring; ms->prev = mm->ring->prev;
      ms->next->prev = ms; ms->prev->next = ms;
   } else {
      ms->next = ms; ms->prev = ms; mm->ring = ms;
   }
   mpx_session_printf(s,
      "Welcome, please tell me your name (just letters).\r\n");
}

static char* help[] = {
   "Either a code using digits 1 to 9 for up to nine colours",
   "or a command may be given:",
   "",
   "colours n       set number of different colours",
   "game            start a new game",
   "help            print this information",
   "len n           set code length",
   "quit            terminate session",
   "summary         print game status",
   "unique          toggle uniqueness",
};

static void request_handler(session* s) {
   mastermind_session* ms = s->handle;
   mastermind* mm = (mastermind*) s->global_handle;
   if (ms->name.len == 0) {
      /* expect name */
      int count = mpx_session_scan(s, &ms->name, &ms->param);
      if (count == 0) {
	 mpx_session_printf(s,
	    "This name is invalid, letters only please.\r\n");
      } else if (count == 1) {
	 /* check that the name is unique */
	 mastermind_session* other = ms->next;
	 while (other != ms && stralloc_diff(&ms->name, &other->name)) {
	    other = other->next;
	 }
	 if (other == ms) {
	    broadcasts(mm, ms, "joins us.");
	    if (mm->game.running) {
	       mpx_session_printf(s, "%.*s\r\n", mm->status.len, mm->status.s);
	    }
	 } else {
	    ms->name.len = 0;
	    mpx_session_printf(s, "This name is already taken.\r\n");
	 }
      } else {
	 close_session(s);
      }
   } else {
      /* participate in a game */
      int count = mpx_session_scan(s, &mm->guess, &ms->param);
      if (count == 0) {
	 mpx_session_printf(s, "Invalid input.\r\n");
      } else if (count == 1) {
	 if (stralloc_diffs(&mm->guess, "quit") == 0) {
	    mpx_session_printf(s, "Bye!\r\n"); close_session(s);
	 } else if (stralloc_diffs(&mm->guess, "game") == 0) {
	    finish_game(mm, ms); start_game(mm);
	    broadcast(mm, ms, &mm->status);
	 } else if (stralloc_diffs(&mm->guess, "help") == 0) {
	    for (unsigned int i = 0; i < sizeof(help)/sizeof(help[0]); ++i) {
	       mpx_session_printf(s, "%s\r\n", help[i]);
	    }
	 } else if (stralloc_diffs(&mm->guess, "summary") == 0) {
	    if (mm->game.running) {
	       mpx_session_printf(s, "%.*s\r\n", mm->status.len, mm->status.s);
	    } else {
	       mpx_session_printf(s, "No game is running.\r\n");
	    }
	 } else if (stralloc_equals(&mm->guess, "unique")) {
	    if (!mm->game.unique && mm->game.colours < mm->game.len) {
	       mpx_session_printf(s,
		  "too few colours to switch to uniqueness\r\n");
	    } else {
	       finish_game(mm, ms);
	       mm->game.unique = !mm->game.unique;
	       mpx_session_printf(s, "S unique = %s\n",
		  mm->game.unique? "true": "false");
	    }
	 } else if (isdigit(mm->guess.s[0])) {
	    if (!mm->game.running) {
	       start_game(mm);
	       broadcast(mm, ms, &mm->status);
	    }
	    unsigned int black; unsigned int white;
	    if (evaluate_guess(&mm->game, &mm->guess, &black, &white)) {
	       mm->line.len = 0;
	       stralloc_cat(&mm->line, &mm->guess);
	       stralloc_cats(&mm->line, "  Black: ");
	       stralloc_catint(&mm->line, black);
	       stralloc_cats(&mm->line, " White: ");
	       stralloc_catint(&mm->line, white);
	       broadcast(mm, ms, &mm->line);
	       stralloc_cats(&mm->status, "\r\n");
	       stralloc_cat(&mm->status, &mm->line);
	       if (black == mm->game.len) {
		  broadcasts(mm, ms, "won the game");
		  mm->game.running = false;
	       }
	    } else {
	       mpx_session_printf(s, "Invalid guess!\r\n");
	    }
	 } else {
	    mpx_session_printf(s, "Invalid command!\r\n");
	 }
      } else if (count == 2) {
	 unsigned int value = get_value(&ms->param);
	 if (stralloc_equals(&mm->guess, "len")) {
	    if (value > 0 && value <= MAX_LENGTH) {
	       if (mm->game.unique && value > mm->game.colours) {
		  mpx_session_printf(s, "length exceeds number of colours\r\n");
	       } else {
		  finish_game(mm, ms);
		  mpx_session_printf(s, "len = %u\n", value);
		  mm->game.len = value;
	       }
	    } else {
	       mpx_session_printf(s, "invalid length value\r\n");
	    }
	 } else if (stralloc_equals(&mm->guess, "colours")) {
	    if (value > 1 && value <= MAX_COLOURS) {
	       if (mm->game.unique && value < mm->game.len) {
		  mpx_session_printf(s,
		     "number of colours must be >= length\r\n");
	       } else {
		  finish_game(mm, ms);
		  mpx_session_printf(s, "colours = %u\n", value);
		  mm->game.colours = value;
	       }
	    } else {
	       mpx_session_printf(s, "F invalid colours value\r\n");
	    }
	 } else {
	    mpx_session_printf(s, "Invalid command!\r\n");
	 }
      } else {
	 close_session(s);
      }
   }
}

static void hangup_handler(session* s) {
   mastermind_session* ms = s->handle;
   if (ms) {
      mastermind* mm = (mastermind*) s->global_handle;
      if (ms->next == ms) {
	 mm->ring = 0;
      } else {
	 broadcasts(mm, ms, "leaves.");
	 ms->next->prev = ms->prev;
	 ms->prev->next = ms->next;
	 if (mm->ring == ms) {
	    mm->ring = ms->next;
	 }
      }
      stralloc_free(&ms->name);
      free(ms);
   }
}

void run_mastermind_service(hostport* hp) {
   srand(getpid() ^ time(0));
   mastermind* mm = calloc(1, sizeof(mastermind));
   if (!mm) return;
   /* default configuration */
   mm->game.len = 4; mm->game.colours = 6; mm->game.unique = true;
   run_mpx_service(hp,
	 "(\\w+|\\d+)"             /* command or guess */
	 "(?:[ \t]+(\\d+))?"       /* optional parameter */
	 "\r?\n",
      open_handler, request_handler, hangup_handler, mm);
}
